#include <string.h>
#include <time.h>
#if defined(HAVE_CLAMAV)
# include <clamav.h>
# ifndef CLAMAV_H
#  define CLAMAV_H
# endif
#endif
#include <gtk/gtk.h>

#include "cfg.h"
#include "win.h"
#include "wincb.h"
#include "windb.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"


/* Utilities */
static gchar *COPY_SHORTEN_STRING(const gchar *s, gint max);
static gchar *FORMATED_NUMBER_STRING(gulong i);
static gchar *TIME_DURATION_STRING(gulong dt, gulong t);

#if defined(CLAMAV_H)
static void WinDBUpdateIterate(
	core_struct *core, win_struct *win,
	GtkCTree *ctree, GtkCTreeNode *pnode, gchar **strv,
	struct cl_node *clam_pnode,
	gboolean verbose, gint *patterns_loaded
);
#endif
void WinDBUpdate(
	win_struct *win, const gchar *db_location,
	gboolean verbose
);
void WinDBQueueUpdate(win_struct *win);
void WinDBClear(win_struct *win);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


#define INSERT_NODE_FOLDER(_s_,_pnode_,_snode_) {       \
 if(columns > 0)                                        \
  strv[0] = (_s_);                                      \
 node = gtk_ctree_insert_node(                          \
  ctree, (_pnode_), (_snode_),                          \
  strv, WIN_LIST_PIXMAP_TEXT_SPACING,                   \
  core->folder_closed_pixmap,                           \
  core->folder_closed_mask,                             \
  core->folder_opened_pixmap,				\
  core->folder_opened_mask,				\
  FALSE, TRUE						\
 );							\
}
#define INSERT_NODE_ITEM(_s_,_pnode_,_snode_)   {       \
 if(columns > 0)                                        \
  strv[0] = (_s_);                                      \
 node = gtk_ctree_insert_node(                          \
  ctree, (_pnode_), (_snode_),                          \
  strv, WIN_LIST_PIXMAP_TEXT_SPACING,                   \
  core->db_item_pixmap, core->db_item_mask,             \
  core->db_item_pixmap, core->db_item_mask,             \
  TRUE, FALSE                                           \
 );                                                     \
}


/*
 *      Returns a copy of the string that is no longer than max
 *      characters with appropriate shortened notation where
 *      appropriate.
 */
static gchar *COPY_SHORTEN_STRING(const gchar *s, gint max)
{
	gint len;

	if(s == NULL)
	    return(NULL);

	len = STRLEN(s);
	if((len > max) && (max > 3))
	{
	    /* Need to shorten string */
	    gint i = len - max + 3;

	    return(g_strdup_printf(
		"...%s", (const gchar *)(s + i)
	    ));
	}
	else
	{
	    return(STRDUP(s));
	}
}

/*
 *	Returns a statically allocated string describing the formatted
 *	number.
 */
static gchar *FORMATED_NUMBER_STRING(gulong i)
{
	gint comma_cnt, slen;
	gchar ss[80], *ss_ptr;
	static gchar ts[80], *ts_ptr;


	g_snprintf(ss, sizeof(ss), "%ld", i);
	slen = STRLEN(ss);                   

	/* 3 digits or less? (no commas needed) */
	if(slen <= 3)
	{
	    strcpy(ts, ss);       
	    return(ts);
	}
	 
	ts_ptr = ts;
	ss_ptr = ss;

	/* Initialize comma counter */
	comma_cnt = slen % 3;
	if(comma_cnt <= 0)
	    comma_cnt = 3;

	/* Iterate through size string until end is reached */
	while(*ss_ptr != '\0')
	{
	    /* Reset comma counter and put in a comma? */
	    if(comma_cnt <= 0)
	    {
		*ts_ptr++ = ',';
		comma_cnt = 3;
	    }

	    *ts_ptr++ = *ss_ptr++;       
	    comma_cnt--;
	}
	 
	/* Null terminate return string */
	*ts_ptr = '\0';

	return(ts);
}

/*
 *      Returns a statically allocated string describing the time
 *      lapsed specified by dt.
 */
static gchar *TIME_DURATION_STRING(gulong dt, gulong t)
{
	gulong ct;
	static gchar buf[80];
			     
	/* Less than one second? */
	if(dt < 1)
	{
	    g_snprintf(
		buf, sizeof(buf),
		"less than a second ago"
	    );
	}
	/* Less than one minute? */
	else if(dt < (1 * 60))  
	{
	    ct = MAX(dt / 1, 1);
	    g_snprintf( 
		buf, sizeof(buf),   
		"%ld second%s ago",
		ct, (ct == 1) ? "" : "s"
	    );
	}     
	/* Less than one hour? */  
	else if(dt < (60 * 60))
	{
	    ct = MAX(dt / 60, 1);
	    g_snprintf(
		buf, sizeof(buf),
		"%ld minute%s ago",
		ct, (ct == 1) ? "" : "s"
	    );
	}     
	/* Less than one day? */
	else if(dt < (24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60, 1);
	    g_snprintf(
		buf, sizeof(buf),
		"%ld hour%s ago",
		ct, (ct == 1) ? "" : "s"
	    );               
	}
	/* Less than one week? */
	else if(dt < (7 * 24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60 / 24, 1);
	    g_snprintf(
		buf, sizeof(buf),
		"%ld day%s ago",
		ct, (ct == 1) ? "" : "s"
	    );
	}     
	/* Less than one month (30 days)? */
	else if(dt < (30 * 24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60 / 24 / 7, 1);
	    g_snprintf(
		buf, sizeof(buf),
		"%ld week%s ago",
		ct, (ct == 1) ? "" : "s"
	    );
	}     
	/* Less than 6 months ago? */
	else if(dt < (6 * 30 * 24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60 / 24 / 30, 1);
	    g_snprintf(
		buf, sizeof(buf),
		"%ld month%s",
		ct, (ct == 1) ? "" : "s"
	    );
	}     
	else
	{
	    gchar *s;
	    time_t tt = (time_t)t;

	    strncpy(buf, ctime(&tt), sizeof(buf));

	    s = strchr(buf, '\n');
	    if(s != NULL)
		*s = '\0';
	}

	buf[sizeof(buf) - 1] = '\0';

	return(buf);
}


#if defined(CLAMAV_H)
/*
 *	Loads Clam AntiVirus child nodes and patterns.
 */
static void WinDBUpdateIterate(
	core_struct *core, win_struct *win,
	GtkCTree *ctree, GtkCTreeNode *pnode, gchar **strv,
	struct cl_node *clam_pnode,
	gboolean verbose, gint *patterns_loaded
)
{
	gint i;
	const struct cli_patt *pattern;
	struct cl_node *clam_cnode;
	GtkCTreeNode *node;
	GtkCList *clist = GTK_CLIST(ctree);
	const gint columns = clist->columns;

	/* Iterate through child clam nodes and add them to the
	 * parent node
	 */
	for(i = 0; i < clam_pnode->nodes; i++)
	{
	    if(win->stop_count > 0)
		break;

	    clam_cnode = clam_pnode->nodetable[i];
	    if(clam_cnode == NULL)
		continue;

	    /* This clam node has no patterns? */
	    if(clam_cnode->list == NULL)
		continue;

	    /* Add node as a folder */
	    INSERT_NODE_FOLDER("", pnode, NULL);
	    if(node == NULL)
		break;

	    /* Add child nodes and items to this node */
	    WinDBUpdateIterate(
		core, win, ctree, node, strv, clam_cnode,
		verbose, patterns_loaded
	    );
	}

	/* Iterate through clam patterns and add them to the parent
	 * node
	 */
	for(pattern = clam_pnode->list;
	    pattern != NULL;
	    pattern = pattern->next
	)
	{
	    if(win->stop_count > 0)
		break;

	    if(STRISEMPTY(pattern->virname))
		continue;

	    /* Add node as an item */
	    INSERT_NODE_ITEM(pattern->virname, pnode, NULL);
	    if(node == NULL)
		break;

	    /* Count this pattern as loaded */
	    *patterns_loaded = (*patterns_loaded) + 1;

	    /* Update progress every 50 patterns loaded */
	    if(verbose && (((*patterns_loaded) % 50) == 0))
		WinStatusProgress(win, -1.0f, TRUE);

	    /* Update status message every 100 patterns loaded */
	    if(verbose && (((*patterns_loaded) % 100) == 0))
	    {
		gchar *patterns_loaded_s = STRDUP(FORMATED_NUMBER_STRING(
		    (gulong)(*patterns_loaded)
		));
		gchar *buf = g_strdup_printf(
		    "%s patterns loaded...",
		    patterns_loaded_s
		);
		WinStatusMessage(win, buf, TRUE);
		g_free(buf);
		g_free(patterns_loaded_s);
	    }

	    /* Thaw and refreeze list every 500 patterns loaded */
	    if(((*patterns_loaded) % 500) == 0)
	    {
		gtk_clist_thaw(clist);
		gtk_clist_freeze(clist);
	    }
	}
}
#endif

/*
 *	Updates the Win's Virus Database Tree with the information
 *	from the specified virus database location.
 */
void WinDBUpdate(
	win_struct *win, const gchar *db_location,
	gboolean verbose
)
{
#if defined(CLAMAV_H)
	struct stat stat_buf;
	struct cl_node *database = NULL;
	gint status, database_entries = 0, patterns_loaded = 0;
#endif
	GtkWidget *w, *toplevel;
	GtkCList *clist;
	const cfg_item_struct *cfg_list;
	core_struct *core;
	GtkCTree *ctree = (win != NULL) ? (GtkCTree *)win->db_ctree : NULL;
	if(ctree == NULL)
	    return;     

	toplevel = win->toplevel;
	clist = GTK_CLIST(ctree);
	core = CORE(win->core);
	cfg_list = core->cfg_list;


	/* Clear */
	WinDBClear(win);


	/* Begin updating */

	/* Reset stop count and update so the stop button can be used */
	win->stop_count = 0;
	WinUpdate(win);

	if(verbose)
	    WinStatusProgress(win, -1.0f, TRUE);

	/* If the virus database location was not specified then use
	 * the one specified in the configuration
	 */
	if(db_location == NULL)
	    db_location = CFGItemListGetValueS(
		cfg_list, CFG_PARM_DB_LOCATION
	    );
#if defined(CLAMAV_H)
	if(db_location == NULL)
	    db_location = cl_retdbdir();
#endif
	if(db_location == NULL)
	{
	    if(verbose)
	    {
		WinStatusMessage(
		    win,
 "No virus database location specified",
		    FALSE
		);
		WinStatusProgress(win, 0.0f, FALSE);
	    }
	    return;
	}


#if defined(CLAMAV_H)
	if(stat(db_location, &stat_buf))
	{
	    if(verbose)
	    {
		WinStatusMessage(     
		    win,
 "Specified virus database does not exist or is not readable",
		    FALSE
		);
		WinStatusProgress(win, 0.0f, FALSE);
	    }
	    return;
	}

	/* Load virus database */
	if(verbose)
	{
	    gchar *buf = g_strdup_printf(
		"Loading virus database \"%s\"...",
		db_location
	    );
	    WinStatusMessage(win, buf, TRUE);
	    g_free(buf);
	}
	if(S_ISDIR(stat_buf.st_mode))
	    status = cl_loaddbdir(db_location, &database, &database_entries);
	else
	    status = cl_loaddb(db_location, &database, &database_entries);
	if(!status && (database != NULL) && (database_entries > 0))
	{
	    gint i;
	    const gint columns = clist->columns;
	    gchar **strv = (gchar **)g_malloc(columns * sizeof(gchar *));
	    GtkCTreeNode *node;

	    if(verbose)
		WinStatusProgress(win, -1.0f, TRUE);

	    for(i = 0; i < columns; i++)
		strv[i] = "";

	    /* Build the virus database */
	    if(cl_buildtrie(database))
	    {
		if(verbose)
		{
		    WinStatusMessage(
			win,
 "Unable to build the virus database",
			FALSE
		    );
		    WinStatusProgress(win, 0.0f, FALSE);
		}
		g_free(strv);
		return;
	    }


	    /* Begin adding the nodes and patterns to the Virus
	     * Database Tree
	     */
	    if(verbose)
		WinStatusProgress(win, -1.0f, TRUE);
	    gtk_clist_freeze(clist);

	    /* Add root node */
	    INSERT_NODE_FOLDER("", NULL, NULL);
	    if(node == NULL)
	    {
		g_free(strv);
		return;
	    }
	    win->db_root_node = node;

	    /* Add all child nodes and items to this node */
	    WinDBUpdateIterate(
		core, win, ctree, node, strv, database, verbose,
		&patterns_loaded
	    );

	    gtk_clist_thaw(clist);
	    if(verbose)
		WinStatusProgress(win, -1.0f, TRUE);


	    /* Delete the virus database */
	    cl_freetrie(database);

	    g_free(strv);
	}
	else
	{
	    WinStatusMessage(
		win,
 "Unable to load the virus database",
		FALSE
	    );
	}
#endif


	/* Update Virus Database Stats Label */
	w = win->db_stats_label;
	if(w != NULL)
	{
	    gulong	cur_t = (gulong)time(NULL),
			last_updated_t = CFGItemListGetValueL(
			    cfg_list, CFG_PARM_DB_LAST_UPDATE
			);
	    gchar	*db_location_s = COPY_SHORTEN_STRING(
			    db_location, 35
			),
			*last_updated_s = STRDUP((last_updated_t > 0) ?
		TIME_DURATION_STRING(cur_t - last_updated_t, last_updated_t) : "Unknown"
			),
			*patterns_loaded_s = STRDUP(FORMATED_NUMBER_STRING(
		(gulong)patterns_loaded
			));
	    gchar *buf = g_strdup_printf(
		"%s\n%s\n%s%s",
		db_location_s,
		last_updated_s,
		patterns_loaded_s,
		(win->stop_count > 0) ? " (Interrupted)" : ""
	    );
	    gtk_label_set_text(GTK_LABEL(w), buf);
	    g_free(buf);
	    g_free(db_location_s);
	    g_free(last_updated_s);
	    g_free(patterns_loaded_s);
	}

	/* Print results */
	if(verbose)
	{
	    gchar *patterns_loaded_s = STRDUP(FORMATED_NUMBER_STRING(
		(gulong)patterns_loaded
	    ));
	    gchar *buf = g_strdup_printf(
		"Virus database loaded with %s patterns",
		patterns_loaded_s
	    );

	    /* Raise Win */
	    WinMap(win);

	    WinStatusMessage(win, buf, FALSE);
	    g_free(buf);
	    g_free(patterns_loaded_s);

	    WinStatusProgress(win, 0.0f, FALSE);
	}
}

/*
 *	Queues a call to WinDBUpdate().
 */
void WinDBQueueUpdate(win_struct *win)
{
	if((win != NULL) ? (win->db_ctree_load_idleid == 0) : FALSE)
	    win->db_ctree_load_idleid = gtk_idle_add_priority(
		G_PRIORITY_LOW,
		WinDBTreeLoadStartIdleCB, win
	    );
}

/*
 *	Deletes all nodes in the Win's Virus Database Tree & Stats
 *	label.
 */
void WinDBClear(win_struct *win)
{
	GtkCList *clist;
	core_struct *core;
	GtkCTree *ctree = (win != NULL) ? (GtkCTree *)win->db_ctree : NULL;
	if(ctree == NULL)
	    return;

	clist = GTK_CLIST(ctree);
	core = CORE(win->core);

	/* Begin clearing the tree */
	gtk_clist_freeze(clist);

	/* Remove the root node */
	if(win->db_root_node != NULL)
	{
	    gtk_ctree_remove_node(ctree, win->db_root_node);
	    win->db_root_node = NULL;
	}

	gtk_clist_thaw(clist);


	/* Clear stats label */
	gtk_label_set_text(GTK_LABEL(win->db_stats_label), "");
}
